/**************************************************************************\
*								 3DPOLY.CPP								   *
*																		   *
*			 Data types and functions for 3D perspective drawing		   *
*								  Polygons								   *
*																		   *
*								Version 1.3b							   *
*																		   *
*							  by Stewart Gordon							   *
*							 July 1998-June 1999						   *
\**************************************************************************/
#include <stdarg.h>
#include <math.h>
#include "3dpoly.h"

// GLOBALS
relpoint clippoint(relpoint, relpoint, float);

// Class polygon: polygon in 3D space

// Constructor: Set all points
polygon::polygon(int NumPoints, int, int, int, ...)
	  : polybase<point>(NumPoints), relp(NULL), scrp(NULL) {
	setdefaultattr();

	va_list ap;

	va_start(ap, NumPoints);
	for(int pno = 0; pno < numpoints; pno++) {
		polypoints[pno].x = va_arg(ap, int);
		polypoints[pno].y = va_arg(ap, int);
		polypoints[pno].z = va_arg(ap, int);
	}
	va_end(ap);
}

// Centroid of polygon, calculated as the mean in each dimension of all
// the points
point polygon::centre() const {
	point cent(0, 0, 0);

	for(int i = 0; i < numpoints; i++) {
		cent.x += polypoints[i].x;
		cent.y += polypoints[i].y;
		cent.z += polypoints[i].z;
	}
	cent.x /= i; cent.y /= i; cent.z /= i;
	return cent;
}

// Determine whether the polygon lies entirely behind the viewer's eye
bool polygon::behindeye(const view& vw) const {
	relpoint vp;
	for(int pno = 0; pno < numpoints; pno++) {
		vp = relpoint(polypoints[pno], vw);
		if(vp.z > 0) return FALSE;
	}
	return TRUE;
}

// Calculate the square of the distance between the centre of the polygon
// and the viewer's eye
float polygon::distsq() const {
	return (centre() - curview->eye).lensq();
}

// Calculate the normal vector to the polygon
vector polygon::normal() const {
	/*	To find the normal vector, it is only necessary to take three
		points, because three points uniquely determine a plane.  It can
		only be assumed that the remaining points lie in the same plane.
		This function works by taking the product of two vectors lying in
		the plane.
	*/
	vector vect1 = polypoints[1] - polypoints[0];
	vector vect2 = polypoints[numpoints - 1] - polypoints[0];
	return (vect1 * vect2);
}

// Calculate the apparent colour of the polygon in the current view
rgb polygon::appcolour(const view& vw) const {

	rgb col = vw.lightcolour
	  * (colour->diffuse * relp->bright(relvector(vw.light, vw))		// diffuse
	  + colour->specular * relp->specbright(relvector(vw.light, vw)))	// specular
	  + vw.ambcolour * colour->diffuse;									// ambient

	return col;
}

// Determine whether the polygon lies in front of or behind a given point
// in the current view
orderrel polygon::relorder(const point& pt, const view& vw) const {
	const float d_plane = eqconst();
	const float d_pt = pt % normal(), d_eye = vw.eye % normal();

	if((d_pt - d_plane) * (d_plane - d_eye) > 0) {	// point and eye on
		return BEHIND;								// opposite side of plane
	}
	else if((d_pt - d_plane) * (d_eye - d_plane) > 0) {	// point and eye on
		return INFRONT;									// same side of plane
	}
	else {				// point on plane or eye on plane
		return EQUAL;
	}
}

// Determine whether the polygon lies in front of or behind a given polygon
// in the current view
orderrel polygon::relorder(const polygon& poly, bool tryother,
	  const view& vw) const {
	int ptsbehind = 0, ptsinfront = 0;

	for(int pno = 0; pno < numpoints; pno++) {
		orderrel pointorder = relorder(poly[pno], vw);
		switch(pointorder) {
			case BEHIND:	ptsbehind++; break;
			case INFRONT:	ptsinfront++;
		}
	}
	if(ptsbehind == 0 && ptsinfront != 0) {	// second shape in front
		return INFRONT;
	}
	else if(ptsbehind != 0 && ptsinfront == 0) {	// second shape behind
		return BEHIND;
	}
	else if(!tryother) {	// shapes coplanar or intersecting
		return EQUAL;
	}
	else {	// if in doubt, try with the shapes the other way round
		return (orderrel)-(poly.relorder(*this, FALSE, vw));
	}
	//return relp->relorder(*poly.relp, tryother);
}

// Render the polygon - convert to viewpoint-relative coordinates, then to
// screen coordinates
void polygon::render(const view& vw) {
	if(relp != NULL) delete relp; relp = NULL;
	if(scrp != NULL) delete scrp; scrp = NULL;
	if(rtdp != NULL) delete rtdp; rtdp = NULL;

	// Convert from absolute to relative coordinates
	relp = new relpoly(this, vw);
	relp->clip();
	if(relp->numpts() == 0) {
		delete relp;
		relp = NULL;
		return;
	}

	// Convert from relative to screen coordinates
	scrp = new scrpoly(relp, vw);
	scrp->clip(-1, -1, 640, 350);
	if(scrp->numpts() == 0) {
		delete scrp;
		scrp = NULL;
		return;
	}

	rtdp = new rtdpoly(scrp);
}

// Draw the polygon
void polygon::draw(const view& vw) {
	if(scrp == NULL) return;

	// set the polygon's line style
	setlinestyle(linestyle, linepattern, linethickness);
	if(filled && colour != NULL && vw.vtype != WIREFRAME) {
		// set the polygon's fill style
		if(vw.vtype == LIGHTSHADE) {
			setfillstyle(fillpattern, vw.palcolour(appcolour(vw)));
		}
		else {
			setfillstyle(fillpattern, fillcolour);
		}

		// draw filled polygon
		rtdp->fill();
	}
	else {
		// draw outline polygon
		rtdp->draw();
	}

}

// Constructor - create viewpoint-relative polygon from absolute polygon
relpoly::relpoly(const polygon* poly, const view& vw)
	  : polybase<relpoint>(poly->numpts()) {
	for(int pno = 0; pno < numpoints; pno++) {
		polypoints[pno] = relpoint((*poly)[pno], vw);
	}

	// close the polygon
	polypoints[numpoints] = polypoints[0];
}

void relpoly::clip(float zmin) {
	// Determine if there is any clipping to do
	for(int pno = 0; pno < numpoints; pno++) {
		if(polypoints[pno].z < zmin)
			break;	// work to do - stop checking
	}
	if(pno == numpoints) return;	// no work to do

	int numnewpts = pno;	// number of points checked so far

	/* Calculate the number of vertices in the clipped polygon
		Each vertex that lies within the range is counted, and a vertex is
		counted where the polygon moves into or out of the range.
	*/
	for(/* continue from previous loop */; pno < numpoints; pno++) {
		// if this vertex and the previous one are either side of the border
		if(pno > 0 && (polypoints[pno].z >= zmin) !=
			  (polypoints[pno - 1].z >= zmin))
			numnewpts++;
		// if the current vertex is in the range
		if(polypoints[pno].z >= zmin)
			numnewpts++;
	}
	// if the first vertex and the last vertex are either side of the border
	if((polypoints[0].z >= zmin) != (polypoints[numpoints - 1].z >= zmin))
		numnewpts++;

	// if the entire polygon is out of range, set polygon to null
	if(numnewpts == 0) {
		numpoints = 0;
		delete[] polypoints;
		polypoints = NULL;
		return;
	}

	// find the first vertex that is in range
	for(pno = 0; pno < numpoints; pno++) {
		if(polypoints[pno].z >= zmin) break;
	}

	// do the clipping
	int pstart = pno, newpno = 0;
	relpoint *newpts = new relpoint[numnewpts + 1];

	for(/* continue from previous loop */; pno < numpoints + pstart; pno++) {
		relpoint &cp = polypoints[pno % numpoints],	// current vertex
		  &pp = polypoints[(pno + numpoints - 1) % numpoints];	// previous vertex
		// add point on edge of range where clipped
		if((cp.z >= zmin) != (pp.z >= zmin)) {
			newpts[newpno++] = clippoint(pp, cp, zmin);
		}
		// add curent point if in range
		if(cp.z >= zmin)
			newpts[newpno++] = cp;
	}

	// close the polygon
	newpts[numnewpts] = newpts[0];

	// assign new polygon to polygon object
	delete[] polypoints;
	polypoints = newpts;
	numpoints = numnewpts;
}

// Calculate the centroid of the polygon
relpoint relpoly::centre() const {
	relpoint v(0, 0, 0);

	for(int pno = 0; pno < numpoints; pno++)
		v += polypoints[pno];

	return v / numpoints;
}

// Calculate the normal vector to the polygon
relvector relpoly::normal() const {
	/*	To find the normal vector, it is only necessary to take three
		points, because three points are always coplanar.  It can only be
		assumed that the remaining points lie in the same plane.
		This function works by taking the product of two vectors lying in
		the plane.
	*/
	/*rel*/vectorbase vect1 = polypoints[1] - polypoints[0];
	/*rel*/vectorbase vect2 = polypoints[numpoints - 1] - polypoints[0];
	return relvector(vect1 * vect2);
}

// Determine whether the polygon lies in front of or behind a given point
orderrel relpoly::relorder(const relpoint& pt) const {
	const float d_plane = eqconst();
	const float d_pt = pt % normal();
	if((d_pt - d_plane) * d_plane > 0) {		// point and eye on
		return BEHIND;							// opposite side of plane
	}
	else if((d_plane - d_pt) * d_plane > 0) {	// point and eye on
		return INFRONT;							// same side of plane
	}
	else {				// point on plane or eye on plane
		return EQUAL;
	}
}

// Determine whether the polygon lies in front of or behind a given polygon
orderrel relpoly::relorder(const relpoly& poly, bool tryother) const {
	int ptsbehind = 0, ptsinfront = 0;

	for(int pno = 0; pno < numpoints; pno++) {
		orderrel pointorder = relorder(poly[pno]);
		switch(pointorder) {
			case BEHIND:	ptsbehind++; break;
			case INFRONT:	ptsinfront++;
		}
	}
	if(ptsbehind == 0 && ptsinfront != 0) {	// second shape in front
		return INFRONT;
	}
	else if(ptsbehind != 0 && ptsinfront == 0) {	// second shape behind
		return BEHIND;
	}
	else if(!tryother) {	// shapes coplanar or intersecting
		return EQUAL;
	}
	else {	// if in doubt, try with the shapes the other way round
		return (orderrel)-(poly.relorder(*this, FALSE));
	}
}

// Calculate the apparent diffuse brightness of the polygon
float relpoly::bright(const relvector& light) const {
/*
	let	l = light direction
		n = plane normal (in direction away from eye)
		s = line of sight
		LD = light density
		AB = apparent brightness

	LD = cos(ln)

	AB = cos(ln)/cos(ns)

	cos(ln) = l.n/|l||n|
	cos(ns) = n.s/|n||s|

	AB = l.n |s|/n.s |l|

*/
	// calculate normal
	relvector norm = normal();
	orderrel order = relorder(polypoints[0] + norm);
	switch(order) {
		case EQUAL: return 0;
		case INFRONT: norm = -norm;	// normal directed away from viewpoint
	}

	// calculate light density
	float dens = (light % norm) / sqrt(light.lensq() * norm.lensq());
	dens = MAX(dens, 0);

	// add ambient light (current directional:ambient ratio = 2:1)
	// commented out - now done in polygon::appcolour
	//dens = (dens + .5) / 1.5;

	// add a slight shine (current matt:gloss ratio: 3:1)
	// (to be replaced with specular lighting component)
	/*relpoint cent = centre();
	float cosva = (norm % cent) / sqrt(norm.lensq() * cent.lensq());
	if(cosva != 0)	// to prevent division by zero
		dens *= (3 + 1 / cosva) / 4;*/

	// ensure value is between 0 and 10
	dens = MAX(dens, 0); dens = MIN(dens, 10);
	return dens;
}

// Calculate the apparent specular brightness of a polygon
float relpoly::specbright(const relvector& light) const {
/*
	let	l = light direction
		n = plane normal (in direction away from eye)
		r = reflected light direction
		s = line of sight
		AB = apparent brightness

	r^ = l^ - 2n^(l.n^)	(v^ = unit vector in direction of v)

	r = l|n|^2 - 2n(l.n)

	AB = cos(rs)

	cos(rs) = r.s/|r||s|
*/

	// calculate normal
	relvector norm = normal();
	orderrel order = relorder(polypoints[0] + norm);
	switch(order) {
		case EQUAL: return 0;
		case INFRONT: norm = -norm;	// normal directed away from viewpoint
	}

	// calculate reflected light direction
	relvector reflect = light * norm.lensq() - norm * (light % norm) * 2;

	// calculate line of sight
	relpoint los = centre();

	// calculate light density
	float dens = -(reflect % los) / sqrt(reflect.lensq() * los.lensq());

	// ensure value is between 0 and 10
	dens = MIN(MAX(dens, 0), 10);
	return dens;
}

// Determine the intercept of a line with a given z-coordinate
relpoint clippoint(relpoint pt1, relpoint pt2, float zmin) {
	relpoint cpoint(0, 0, zmin);

	/*	xm = (x2-x1)/(z2-z1)
		x-x1 = xm(z-z1)
		x = xm(z-z1)+x1

		ym and y are calculated similarly

		Division by zero will not occur, because this function is called
		only when pt1.z and pt2.z differ.
	*/
	float m = (pt1.x - pt2.x) / (pt1.z - pt2.z);
	cpoint.x = m * (zmin - pt1.z) + pt1.x;
	m = (pt1.y - pt2.y) / (pt1.z - pt2.z);
	cpoint.y = m * (zmin - pt1.z) + pt1.y;

	return cpoint;
}

