/**************************************************************************\
*								3DVECTOR.CPP							   *
*																		   *
*			 Data types and functions for 3D perspective drawing		   *
*						  Vectors, points and view						   *
*																		   *
*								Version 1.3a							   *
*																		   *
*							  by Stewart Gordon							   *
*							 July 1998-May 1999							   *
\**************************************************************************/
#include <math.h>
#include <stdarg.h>
#include <graphics.h>
#include <mem.h>
#include "3dvector.h"
#include "scrpoly.h"
#define PI			acos(-1)

// GLOBALS
view *curview = NULL;
relpoint clippoint(relpoint, relpoint, float);

// Class vectorbase: 3D vectorbase

// Add one vector to another
vectorbase vectorbase::operator+(vectorbase a) const {
	vectorbase value;
	value.x = x + a.x;
	value.y = y + a.y;
	value.z = z + a.z;
	return value;
}

// Subtract one vector from another
vectorbase vectorbase::operator-(vectorbase a) const {
	vectorbase value;
	value.x = x - a.x;
	value.y = y - a.y;
	value.z = z - a.z;
	return value;
}

// Return the negative of the vector
vectorbase vectorbase::operator-() const {
	vectorbase value;
	value.x = -x;
	value.y = -y;
	value.z = -z;
	return value;
}

// Vector product of two vectors
vectorbase vectorbase::operator*(vectorbase a) const {
// = i (y1 z2 - z1 y2) + j (z1 x2 - x1 z2) + k (x1 y2 - y1 x2)
	vectorbase value;
	value.x = y * a.z - z * a.y;
	value.y = z * a.x - x * a.z;
	value.z = x * a.y - y * a.x;
	return value;
}

// Scalar product of two vectors
float vectorbase::operator%(vectorbase a) const {
	return x * a.x + y * a.y + z * a.z;
}

// Square of the distance of the point from the origin
// Used as a vector, this is the square of the magnitude
float vectorbase::lensq() const {
	return x * x + y * y + z * z;
}


// Class point: point in 3D space

// X-coordinate of a point on the screen
float point::screenx(const view& vw) const {
	return relpoint((point)*this, vw).screenx(vw);
}

float point::screeny(const view& vw) const {
	return relpoint((point)*this, vw).screeny(vw);
}

// Class relpoint: position of a point relative to a particular view

// Constructor: relative position of given vector
relvector::relvector(vector pt, const view& vw) : vectorbase(pt) {
	double azim = rad(vw.azim), alt = rad(vw.alt);
	if(azim != 0)
		*this = relvector(x * cos(azim) + z * sin(azim),
		  y, z * cos(azim) - x * sin(azim));
	if(alt != 0)
		*this = relvector(x, y * cos(alt) - z * sin(alt),
		  z * cos(alt) + y * sin(alt));
}

// Constructor: relative position of given point
relpoint::relpoint(point pt, const view& vw) {
	*(relvector*)this = relvector((vector)(pt - vw.eye), vw);
}

// Add a vector to a point
relpoint relpoint::operator+(relvector a) const {
	relpoint value;
	value.x = x + a.x;
	value.y = y + a.y;
	value.z = z + a.z;
	return value;
}

// Subtract a vector from a point
relpoint relpoint::operator-(relvector a) const {
	relpoint value;
	value.x = x - a.x;
	value.y = y - a.y;
	value.z = z - a.z;
	return value;
}

// Subtract a point from a point
relvector relpoint::operator-(relpoint a) const {
	relvector value;
	value.x = x - a.x;
	value.y = y - a.y;
	value.z = z - a.z;
	return value;
}

// X-coordinate of a point on the screen
float relpoint::screenx(const view& vw) const {
	if(z > 0) {
		return vw.vpx + x / z * vw.xscale;
	}
	else {
		return vw.vpx + x * vw.xscale;
	}
}

// Y-coordinate of a point on the screen
float relpoint::screeny(const view& vw) const {
	if(z > 0) {
		/*	Unlike in screenx, the y-displacement of the point from the
			vanishing point is subtracted from the y-coordinate of the
			vanishing point, not added.  This is because in the device
			coordinate system the y-coordinate increases downwards, and in
			our coordinate system it is to increase upwards.
		*/
		return vw.vpy - y / z * vw.yscale;
	}
	else {
		return vw.vpy - y * vw.yscale;
	}
}

// Class view: view of 3D space

// Copy data from one view into another
const view& view::operator=(const view& vw) {
	delete pal;
	memcpy(this, &vw, sizeof(view));
	pal = NULL;
	return *this;
}

// Move forward in 2 dimensions (x and z but not y)
void view::fwd(float d) {
	double razim = rad(azim);

	eye.x -= d * sin(razim);
	eye.z += d * cos(razim);
}

// Move forward in all 3 dimensions
void view::fwd3(float d) {
	double razim = rad(azim), ralt = rad(alt);

	eye.x -= d * cos(ralt) * sin(razim);
	eye.y += d * sin(ralt);
	eye.z += d * cos(ralt) * cos(razim);
}

void view::initpal(int pmin, int pmax) {
	if(pal != NULL) delete pal;
	pal = new palette(pmin, pmax);
}

