/**************************************************************************\
*								 SCRPOLY.CPP							   *
*																		   *
*			 Data types and functions for 3D perspective drawing		   *
*				   Mapping to screen coordinates and clipping			   *
*																		   *
*								 Version 1.1							   *
*																		   *
*							  by Stewart Gordon							   *
*							  July-August 1998							   *
\**************************************************************************/
#include <stdio.h>
#include <miscids.h>
#include "3dpoly.h"
#include "scrpoly.h"

// GLOBALS
float clipval(scrpoint, scrpoint, int, dimension);
bool inrange(float, int, clipside);

// Class scrpoly: polygon in screen coordinates

// Constructor: create screen polygon from viewpoint-relative polygon
scrpoly::scrpoly(const relpoly* poly, const view& vw)
	  : polybase<scrpoint>(poly->numpts()) {
	for(int pno = 0; pno <= numpoints; pno++) {
		polypoints[pno].x = (*poly)[pno].screenx(vw);
		polypoints[pno].y = (*poly)[pno].screeny(vw);
	}
}

// Clip the polygon to specified boundaries
void scrpoly::clip(int left, int top, int right, int bottom) {
	clipedge(left, MINVAL, X);
	clipedge(right, MAXVAL, X);
	clipedge(top, MINVAL, Y);
	clipedge(bottom, MAXVAL, Y);
}

// Clip the polygon to a specific edge
void scrpoly::clipedge(int border, clipside cside, dimension dim) {
	float scrpoint::*td = (dim == X) ? &scrpoint::x : &scrpoint::y;
	float scrpoint::*od = (dim == Y) ? &scrpoint::x : &scrpoint::y;

	// Determine if there is any clipping to do
	for(int pno = 0; pno < numpoints; pno++) {
		if(!inrange(polypoints[pno].*td, border, cside))
			break;	// work to do - stop checking
	}
	if(pno == numpoints) return;	// no work to do

	int numnewpts = pno;	// number of points checked so far

	/* Calculate the number of vertices in the clipped polygon
		Each vertex that lies within the range is counted, and a vertex is
		counted where the polygon moves into or out of the range.
	*/
	for(/* continue from previous loop */; pno < numpoints; pno++) {
		// if this vertex and the previous one are either side of the border
		if(pno > 0 && inrange(polypoints[pno].*td, border, cside) !=
			  inrange(polypoints[pno - 1].*td, border, cside))
			numnewpts++;
		// if the current vertex is in the range
		if(inrange(polypoints[pno].*td, border, cside))
			numnewpts++;
	}
	// if the first vertex and the last vertex are either side of the border
	if(inrange(polypoints[0].*td, border, cside) !=
		  inrange(polypoints[numpoints - 1].*td, border, cside))
		numnewpts++;

	// if the entire polygon is out of range, set polygon to null
	if(numnewpts == 0) {
		numpoints = 0;
		delete[] polypoints;
		polypoints = NULL;
		return;
	}

	// find the first vertex that is in range
	for(pno = 0; pno < numpoints; pno++) {
		if(inrange(polypoints[pno].*td, border, cside)) break;
	}

	// do the clipping
	int pstart = pno, newpno = 0;
	scrpoint *newpts = new scrpoint[numnewpts + 1];

	for(/* continue from previous loop */; pno < numpoints + pstart; pno++) {
		scrpoint &cp = polypoints[pno % numpoints],	// current vertex
		  &pp = polypoints[(pno + numpoints - 1) % numpoints];	// previous vertex
		// add point on edge of range where clipped
		if(inrange(cp.*td, border, cside) != inrange(pp.*td, border, cside)) {
			newpts[newpno].*td = border;
			newpts[newpno++].*od = clipval(pp, cp, border, dim);
		}
		// add curent point if in range
		if(inrange(cp.*td, border, cside))
			newpts[newpno++] = cp;
	}

	// close the polygon
	newpts[numnewpts] = newpts[0];

	// assign new polygon to polygon object
	delete[] polypoints;
	polypoints = newpts;
	numpoints = numnewpts;
}

// Determine whether a point is outside the clipping range
bool inrange(float coord, int border, clipside cside) {
	return (cside == MINVAL && coord >= border)
	  || (cside == MAXVAL && coord <= border);
}

// Determine the y-coordinate of the intercept of a line with a given
// x-coordinate
float clipval(scrpoint pt1, scrpoint pt2, int xborder, dimension dim) {
	float scrpoint::*td = (dim == X) ? &scrpoint::x : &scrpoint::y;
	float scrpoint::*od = (dim == Y) ? &scrpoint::x : &scrpoint::y;

	/*	m = (y2-y1)/(x2-x1)
		y-y1 = m(x-x1)
		y = m(x-x1)+y1

		Division by zero will not occur, because this function is called
		only when pt1.x and pt2.x differ.
	*/
	float m = (pt1.*od - pt2.*od) / (pt1.*td - pt2.*td);
	return m * (xborder - pt1.*td) + pt1.*od;
}

// Convert array of points to integers
int* scrpoly::points() {
	int *pts = new int[numpoints * 2 + 2];
	for(int pno = 0; pno <= numpoints; pno++) {
		pts[pno * 2] = polypoints[pno].x;
		pts[pno * 2 + 1] = polypoints[pno].y;
	}
	return pts;
}

// Class rtdpoly: polygon in screen coordinates as integers (ready to draw)
// Constructor: Create polygon from screen polygon
rtdpoly::rtdpoly(const scrpoly* poly)
	  : polybase<intpoint>(poly->numpts()) {
	for(int pno = 0; pno <= numpoints; pno++) {
		polypoints[pno].x = (*poly)[pno].x;
		polypoints[pno].y = (*poly)[pno].y;
	}
}

